/* -*- Mode: C++ -*-
   vi:si:et:sw=4:sts=4:ts=4
 */
/*
  kerbsystray - A systray util to list kerberos tickets on Windows
  Copyright (C) 2007 Michele Baldessari <michele@pupazzo.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/

#include "kerbutil.h"

char *char_to_hex(const char *str)
{
	char *tmp = NULL, *p;
	size_t size;
	unsigned int i;
	const char hex[] = "0123456789ABCDEF";

	size = (strlen(str) * 2) + 1;
	tmp = (char *)malloc(size);
	memset(tmp, 0, size);

	i = 0;
	p = (char *)str;
	while (*p && *p != '\0' && i < size) {
		tmp[i++] = hex[(*p >> 4) & 0xf];
		tmp[i++] = hex[*p & 0xf];
		p++;
	}
	tmp[i] = '\0';
	return tmp;
}

KerberosUtilClass::KerberosUtilClass(void)
{
    LSA_STRING Name;
    NTSTATUS Status;

    Initialized = FALSE;
    Status = LsaConnectUntrusted (&LsaLogon);
    if (FAILED(Status)) 
        return;

    Name.Buffer = MICROSOFT_KERBEROS_NAME_A;
    Name.Length = (USHORT)strlen(Name.Buffer);
    Name.MaximumLength = Name.Length + 1;

    Status = LsaLookupAuthenticationPackage(LsaLogon, &Name, &AuthPackage);
    if (FAILED(Status)) 
        return;
	
    pTicketCacheResponse = NULL;
    Initialized = TRUE;
    return;
}

KerberosUtilClass::~KerberosUtilClass(void)
{
	if (!Initialized)
		return;
	CloseHandle(LsaLogon);
	if (pTicketCacheResponse != NULL) {
		LsaFreeReturnBuffer(pTicketCacheResponse);
		pTicketCacheResponse = NULL;
	}
	return;
}

BOOL 
KerberosUtilClass::isInitialized()
{
	return Initialized;
}

BOOL 
KerberosUtilClass::purgeAllTickets()
{
    NTSTATUS Status = 0;
    NTSTATUS SubStatus = 0;
    KERB_PURGE_TKT_CACHE_REQUEST PurgeRequest;
	
	if (!Initialized)
		return FALSE;

    PurgeRequest.MessageType = KerbPurgeTicketCacheMessage;
    PurgeRequest.LogonId.LowPart = 0;
    PurgeRequest.LogonId.HighPart = 0;
    PurgeRequest.ServerName.Buffer = L"";
    PurgeRequest.ServerName.Length = 0;
    PurgeRequest.ServerName.MaximumLength = 0;
    PurgeRequest.RealmName.Buffer = L"";
    PurgeRequest.RealmName.Length = 0;
    PurgeRequest.RealmName.MaximumLength = 0;
    Status = LsaCallAuthenticationPackage(LsaLogon,
                                          AuthPackage,
                                          &PurgeRequest,
                                          sizeof(PurgeRequest),
                                          NULL,
                                          NULL,
                                          &SubStatus
                                          );
    if (FAILED(Status) || FAILED(SubStatus))
        return FALSE;
	/* Kerberos Cache is empty at this point anyway */
	if (pTicketCacheResponse != NULL) {
		LsaFreeReturnBuffer(pTicketCacheResponse);
		pTicketCacheResponse = NULL;
	}
	return TRUE;
}

BOOL 
KerberosUtilClass::refreshKerberosTicketCache ()
{
	NTSTATUS Status = 0;
	NTSTATUS SubStatus = 0;
    ULONG RequestSize, ResponseSize;
    PKERB_QUERY_TKT_CACHE_REQUEST pTicketCacheRequest = NULL;

	if (!Initialized)
		return FALSE;

	if (pTicketCacheResponse != NULL) {
		LsaFreeReturnBuffer(pTicketCacheResponse);
		pTicketCacheResponse = NULL;
	}

    RequestSize = sizeof(*pTicketCacheRequest) + 1;
    pTicketCacheRequest = (PKERB_QUERY_TKT_CACHE_REQUEST) LocalAlloc(LMEM_ZEROINIT, RequestSize);
    if (!pTicketCacheRequest) {
        CloseHandle(LsaLogon);
        return NULL;
    }

    pTicketCacheRequest->MessageType = KerbQueryTicketCacheMessage;
    pTicketCacheRequest->LogonId.LowPart = 0;
    pTicketCacheRequest->LogonId.HighPart = 0;
    
    Status = LsaCallAuthenticationPackage(LsaLogon, AuthPackage, pTicketCacheRequest,
		RequestSize, (PVOID *)&pTicketCacheResponse, &ResponseSize, &SubStatus);   

    if (Status != 0 || SubStatus != 0)
	    return FALSE;
	return TRUE;
}

unsigned int 
KerberosUtilClass::getTicketCacheCount ()
{
	if (pTicketCacheResponse == NULL)
		return -1;
	return pTicketCacheResponse->CountOfTickets;
}

BOOL 
KerberosUtilClass::getTicketFromCache (unsigned int ticket_nr, PKERB_TICKET_CACHE_INFO ticket)
{
	if (pTicketCacheResponse == NULL || ticket_nr < 0 || 
		ticket_nr >= pTicketCacheResponse->CountOfTickets)
		return FALSE;

	*ticket = pTicketCacheResponse->Tickets[ticket_nr];
	return TRUE;
}

BOOL 
KerberosUtilClass::getMSCacheTicketFromCacheInfo(unsigned int ticket_nr, PKERB_EXTERNAL_TICKET *ticket)
{
    NTSTATUS Status = 0;
    NTSTATUS SubStatus = 0;
    ULONG RequestSize;
    PKERB_RETRIEVE_TKT_REQUEST pTicketRequest = NULL;
    PKERB_RETRIEVE_TKT_RESPONSE pTicketResponse = NULL;
	PKERB_TICKET_CACHE_INFO pTicketInfo = NULL;
    ULONG ResponseSize;

	if (!Initialized || ticket_nr >= getTicketCacheCount())
		return FALSE;

	pTicketInfo = &(pTicketCacheResponse->Tickets[ticket_nr]);
    RequestSize = sizeof(*pTicketRequest) + pTicketInfo->ServerName.Length;
    pTicketRequest = (PKERB_RETRIEVE_TKT_REQUEST) LocalAlloc(LMEM_ZEROINIT, RequestSize);
    if (!pTicketRequest)
        return FALSE;

    pTicketRequest->MessageType = KerbRetrieveEncodedTicketMessage;
    pTicketRequest->LogonId.LowPart = 0;
    pTicketRequest->LogonId.HighPart = 0;
    pTicketRequest->TargetName.Length = pTicketInfo->ServerName.Length;
    pTicketRequest->TargetName.MaximumLength = pTicketInfo->ServerName.Length;
    pTicketRequest->TargetName.Buffer = (PWSTR) (pTicketRequest + 1);
    memcpy(pTicketRequest->TargetName.Buffer,pTicketInfo->ServerName.Buffer, pTicketInfo->ServerName.Length);
    pTicketRequest->CacheOptions = 0;
    pTicketRequest->CacheOptions |= KERB_RETRIEVE_TICKET_CACHE_TICKET;
    pTicketRequest->EncryptionType = pTicketInfo->EncryptionType;
    pTicketRequest->TicketFlags = 0;

    Status = LsaCallAuthenticationPackage(LsaLogon, AuthPackage, pTicketRequest,
		RequestSize, (PVOID *)&pTicketResponse, &ResponseSize, &SubStatus);

    LocalFree(pTicketRequest);
    if (FAILED(Status) || FAILED(SubStatus))
        return FALSE;
    
    /* otherwise return ticket */
    *ticket = &(pTicketResponse->Ticket);
    if ( pTicketInfo->TicketFlags & KERB_TICKET_FLAGS_initial )
        (*ticket)->TicketFlags |= KERB_TICKET_FLAGS_initial;

	return TRUE;
}